#! /usr/bin/env python3

"""Scans helm charts for policy violations and container images.

Renders helm charts off of helm template files, produced by the Grand Central
pipeline, using the application name and environment as input.

For each environment, scans the helm chart for violations, searches for images
and sends the images to cbctl.

Produces an HTML report containing the policy violations. This file is added
to a final report that contains image vulnerability information.
"""

import os
import subprocess
import yaml
import shutil
import hashlib
import json
from jinja2 import Environment, FileSystemLoader

CBC_ENVS = ["prd1ew2", "int2use2", "int1usw2"]
K8S_SCAN_ENABLED = os.getenv("K8S_SCAN", None).lower() == "true"
APP_NAME = os.getenv("APP_NAME", None)
GIT_USER = os.getenv("GIT_USER", None)
GIT_TOKEN = os.getenv("GIT_TOKEN", None)
REPO_URL = f"https://{GIT_USER}:{GIT_TOKEN}@gitlab.bit9.local/mcs2/grand-central.git"
REP_PATH_CLI = os.getenv("REP_PATH_CLI", ".")
JINJA_REPORT_TEMPLATE = "kubernetes/report_template.jinja2.html"


def validate_variables():
    """Validate the presence of required environment variables.

    Raises:
        ValueError: If any required environment variable is not set.
    """

    required_vars = ["APP_NAME", "GIT_USER", "GIT_TOKEN"]

    for var in required_vars:
        if not os.environ.get(var):
            raise ValueError(f"** ERROR: {var} environment variable is not set.")


def run_cmd(command, capture_json=False):
    """Run a shell command and capture its output.

    Args:
        command (list): The command to be executed.
        capture_json (bool): Whether to capture JSON output.

    Returns:
        tuple: A tuple containing the return code and output.
    """

    try:
        process = subprocess.run(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            check=False,
        )
        output = process.stdout.strip()

        if capture_json:
            try:
                json.loads(output)
            except json.JSONDecodeError:
                print("** ERROR: Failed to extract valid JSON from the command output.")
                exit(0)

        return process.returncode, output
    except subprocess.CalledProcessError as e:
        print(f"** ERROR: Command execution failed: {e}")
        exit(0)


def render_helm_chart(APP_NAME: str, env: str):
    """Render the Helm chart template for a given application and environment.

    Args:
        APP_NAME (str): The application name.
        env (str): The environment.

    Returns:
        str: The path to the rendered Helm chart template.
    """

    original_dir = os.getcwd()
    os.environ["ENV"] = env
    os.environ["RELEASES"] = APP_NAME

    os.chdir("grand-central")

    print(f"** INFO: Rendering the helm chart template for {APP_NAME} in {env} ..")
    run_cmd(["make", "template"])

    os.chdir(original_dir)

    _helm_chart = f"builds/{env}/{APP_NAME}/helmfile_log/helmfile_template"

    return _helm_chart


def extract_images(chart_data: yaml.YAMLObject):
    """Extract container images from Helm chart data.

    Args:
        chart_data (yaml.YAMLObject): YAML object representing the Helm chart data.

    Returns:
        list: A list of unique container images.
    """

    images = []
    containers = []

    for doc in chart_data:
        for resource in doc:
            if "kind" in resource:
                if doc[resource] in ["Deployment", "StatefulSet", "Job"]:
                    containers = (
                        doc.get("spec", {})
                        .get("template", {})
                        .get("spec", {})
                        .get("containers", [])
                    )
                elif doc[resource] == "Pod":
                    if "initContainers" in doc["spec"]:
                        containers = doc.get("spec", {}).get("initContainers", [])
                    elif "containers" in doc["spec"]:
                        containers = doc.get("spec", {}).get("containers", [])

                for container in containers:
                    image = container.get("image", "")

                    if image and image not in images:
                        images.append(image)

    return images


def set_text_color(severities: any):
    """Determine the text color based on severity levels.

    Args:
        severities (Union[str, List[str]]): Severity level(s).

    Returns:
        str: The color code.
    """

    color_return = "dark"
    color_risk = {
        "critical": "danger",
        "medium": "warning",
        "low": "primary",
        "pass": "success",
    }
    severity_order = {"critical": 3, "medium": 2, "low": 1, "pass": 0}

    if isinstance(severities, str):
        color_return = color_risk.get(severities.lower(), "dark")
    elif isinstance(severities, list):
        valid_severities = []

        for severity in severities:
            severity = str(severity).lower()

            if severity in color_risk:
                valid_severities.append(severity)

        if valid_severities:
            highest_severity = max(valid_severities, key=lambda x: severity_order[x])

            color_return = color_risk[highest_severity]

    return color_return


def generate_violations_report(
    REP_PATH_CLI, APP_NAME, JINJA_REPORT_TEMPLATE, k8s_objects=None, env=None
):
    """Generate HTML report containing policy violations.

    Args:
        REP_PATH_CLI (str): Path to the CLI report.
        APP_NAME (str): Application name.
        JINJA_REPORT_TEMPLATE (str): Path to the Jinja2 report template.
        k8s_objects (dict): Kubernetes objects data.
        env (str): Environment.

    Returns:
        None
    """

    k8s_final_report = f"{REP_PATH_CLI}/kubernetes_{APP_NAME}_{env}_final_report.html"
    jinja_env = Environment(loader=FileSystemLoader("."))
    template = jinja_env.get_template(JINJA_REPORT_TEMPLATE)
    item_id = hashlib.sha256(os.urandom(16)).hexdigest()
    severities = []

    headers = [
        "Namespace",
        "Kind",
        "Name",
        "Policy",
        "Rule",
        "Risk",
        "File",
    ]

    if len(k8s_objects) > 0:
        rows = []

        for _obj in k8s_objects["objects"]:
            if len(_obj.get("policy_violations", [])) > 0:
                for v in _obj["policy_violations"]:
                    namespace = _obj.get("namespace", "UNKNOWN")
                    kind = _obj["kind"]
                    name = _obj["name"]
                    file_path = _obj["file_path"]
                    policy_name = _obj["policy"]
                    violation_rule = v["rule"]
                    violation_risk = v["risk"]

                    violation = [
                        namespace,
                        kind,
                        name,
                        policy_name,
                        violation_rule,
                        violation_risk,
                        file_path,
                    ]

                    rows.append(violation)
                    severities.append(violation_risk)
    else:
        rows.append(["No violations found", "", "", "", "", "", ""])

    header_color = set_text_color(severities)

    rendered_html = template.render(
        headers=headers,
        header_color=header_color,
        rows=rows,
        item_id=item_id,
        app_name=APP_NAME,
        env=env,
    )

    with open(k8s_final_report, "w") as output_file:
        output_file.write(rendered_html)


def main():
    """Main function to initiate scanning activities."""

    validate_variables()

    if not K8S_SCAN_ENABLED:
        print("** ERROR: Kubernetes helm chart scan not enabled. Skipping scan ..")
        exit(0)

    print(f"** INFO: Initiating scanning activities for {APP_NAME} ..")
    print("** INFO: Cloning the Grand-Central repo ..")

    # Hardcoded the branch here for testing and until amanno@vmware.com removes the cluster check in prod
    run_cmd(["git", "clone", "-b", "remove-template-cluster-check", f"{REPO_URL}"])

    for env in CBC_ENVS:
        rendered_helm_chart = render_helm_chart(APP_NAME, env)

        if not rendered_helm_chart.endswith((".yml", ".yaml")):
            try:
                helm_chart = shutil.copy(rendered_helm_chart, "helm_chart.yml")
            except FileNotFoundError:
                print(f"** ERROR: The file '{helm_chart}' could not be found.")
                exit(0)

            print("** INFO: Successfully renamed the helm chart file.")

        abs_helm_chart = os.path.join(os.getcwd(), helm_chart)

        validate_k8s_object_cmd = [
            "cbctl",
            "k8s-object",
            "validate",
            "-f" f"{abs_helm_chart}",
            "-o",
            "json",
        ]

        print(f"** INFO: Initiating helm chart scan for {env} ..")
        return_code, tmp_json = run_cmd(validate_k8s_object_cmd, capture_json=True)

        if return_code == 0 or return_code == 50:
            k8s_objects_json = json.loads(tmp_json)

            print(f"** INFO: Generating kubernetes violations report for {env} ..")
            generate_violations_report(
                REP_PATH_CLI, APP_NAME, JINJA_REPORT_TEMPLATE, k8s_objects_json, env
            )

        print(f"** INFO: Searching for images in helm chart for {env} ..")

        with open(helm_chart, "r") as stream:
            chart_data = list(yaml.load_all(stream, Loader=yaml.FullLoader))

        images = extract_images(chart_data)

        if images:
            print(f"** INFO: Container images detected!")
            print(f"** INFO: Initiating CBCTL scans for the following images in {env}:")
            print("\n".join(images))

            os.environ["SCAN_IMAGE_LIST"] = " ".join(images)

            run_cmd(["sh", "scan_cbctl.sh"])

    print(f"** INFO: All scanning activities are now complete.")


if __name__ == "__main__":
    main()
