#!/bin/sh

# disable variables not being referenced warning
# shellcheck disable=SC2034

# Logs into BCOM artifactory if the image is hosted there
mkdir -p ~/.docker && echo "${CB_DOCKER_AUTH_CONFIG}" | base64 -d >~/.docker/config.json

# Check if we are scanning an RE image
# check_is_re_image target_image
check_is_re_image() {
    IMAGE="$1"
    # Get image name without the tag
    IMAGE_NAME=$(echo "${IMAGE}" | sed 's|:[^:/]*$||')
    if echo "${IMAGE_NAME}" | grep "re/releng" >/dev/null 2>&1; then
        # IMAGE_NAME is an RE image (not considering tags)
        return 0
    fi

    return 1
}

# check_image_base target_image base_image
# check_image_base artifactory-pub.bit9.local:5002/re/releng/photon3-go:220502_0908 artifactory-pub.bit9.local:5001/re/releng/photon3:220502_0714
check_image_base() {
    BASE_LAYER_CHAIN=$(docker inspect --format='{{.RootFS.Layers}}' "${2}" | tr -d '[],')
    TARGET_LAYER_CHAIN=$(docker inspect --format='{{.RootFS.Layers}}' "${1}" | tr -d '[],')
    if test -z "${BASE_LAYER_CHAIN}" || test -z "${TARGET_LAYER_CHAIN}"; then
        # We could not get the layers for some reason
        return 2
    fi
    # check whether the chain of base layers is contained within the target layers and return
    case "$TARGET_LAYER_CHAIN" in
    *"$BASE_LAYER_CHAIN"*)
        return 0
        ;;
    esac

    # fail if base is not a substring of target
    return 1
}

# error_msg
error_msg() {
    printf "\u001b[31;1m"
    printf "%s" "$*"
    printf "\u001b[0m\n"
}

# cleanup between images
cleanup() {
    rm "${REP_PATH_CLI}"/base_image*.json 2>/dev/null || true
    rm "${REP_PATH_CLI}"/target_image*.json 2>/dev/null || true
}

# Function to login into AWS ECR
login_to_ecr() {
    # Temporarily disable debugging to avoid leaking the creds
    set +x

    IMAGE_NAME="$1"
    aws_account=$(echo "${IMAGE_NAME}" | cut -d'.' -f1)
    region=$(echo "${IMAGE_NAME}" | awk -F'\\.|/' '{print $4}')

    # Load the psec AWS credentials
    export AWS_ACCESS_KEY_ID="${AWS_SECPROD_KEY_ID}"
    export AWS_SECRET_ACCESS_KEY="${AWS_SECPROD_SECRET}"
    psec_role_arn="arn:aws:iam::${AWS_SECPROD_ACCOUNT_ID}:role/psec-ecr-image-scanner"
    target_account_role_arn="arn:aws:iam::${aws_account}:role/psec-ecr-image-scanner"

    # Assume the psec role in SECPROD
    secprod_token_json=$(aws sts assume-role --role-arn "${psec_role_arn}" --role-session-name ecr-session)

    # Load the credentials for the psec role in SECPROD
    AWS_ACCESS_KEY_ID=$(echo "${secprod_token_json}" | jq -r '.Credentials''.AccessKeyId')
    export AWS_ACCESS_KEY_ID

    AWS_SECRET_ACCESS_KEY=$(echo "${secprod_token_json}" | jq -r '.Credentials''.SecretAccessKey')
    export AWS_SECRET_ACCESS_KEY

    AWS_SESSION_TOKEN=$(echo "${secprod_token_json}" | jq -r '.Credentials''.SessionToken')
    export AWS_SESSION_TOKEN

    # Assume the psec role in the target AWS account
    target_token_json=$(aws sts assume-role --role-arn "${target_account_role_arn}" --role-session-name ecr-session)

    # Load the credentials for the psec role in the target AWS account
    AWS_ACCESS_KEY_ID=$(echo "${target_token_json}" | jq -r '.Credentials''.AccessKeyId')
    export AWS_ACCESS_KEY_ID

    AWS_SECRET_ACCESS_KEY=$(echo "${target_token_json}" | jq -r '.Credentials''.SecretAccessKey')
    export AWS_SECRET_ACCESS_KEY

    AWS_SESSION_TOKEN=$(echo "${target_token_json}" | jq -r '.Credentials''.SessionToken')
    export AWS_SESSION_TOKEN

    # Get the ECR credentials for the target AWS account
    docker_password=$(aws ecr get-login-password --region "${region}")

    # Log into ECR in the target AWS account
    docker login --username AWS --password "${docker_password}" "${aws_account}.dkr.ecr.${region}.amazonaws.com"

    unset AWS_ACCESS_KEY_ID
    unset AWS_SECRET_ACCESS_KEY
    unset AWS_SESSION_TOKEN

    # Re-enable debugging
    set -x
}

set_image_vars() {
    # Take the first argument as the IMAGE
    IMAGE_ARG=$1

    IMAGE_REGISTRY_NAME=$(echo "${IMAGE_ARG}" | cut -d'/' -f 1)

    IMAGE_REPO_NAME=$(echo "${IMAGE_ARG}" | cut -d'/' -f 2- | cut -d':' -f1 | sed 's:/:_:g')

    IMAGE_TAG=$(echo "${IMAGE_ARG}" | rev | cut -d":" -f1 | rev)

    RES_CLI="${REP_PATH_CLI}/${SCANNER}__${CI_PIPELINE_ID}__${JENKINS_BUILD_ID}__${IMAGE_REGISTRY_NAME}__${IMAGE_REPO_NAME}__${IMAGE_TAG}"

    JSON_OUTPUT="${REP_PATH_CLI}/psec__${CI_PIPELINE_ID}__${JENKINS_BUILD_ID}__${SCANNER}__${IMAGE_REGISTRY_NAME}__${IMAGE_REPO_NAME}__${IMAGE_TAG}.json"
}

# Get info from `/etc/os-release`
get_info_os() {
    # Take the first argument as the IMAGE
    IMAGE="${1}"

    # Distinguishable from what docker run might error out with
    EXIT_NOSHELL=129

    if ! docker run --rm --entrypoint /bin/sh "${IMAGE}" -c "true"; then
        echo "Could not run shell in ${IMAGE}" >&2
        return "${EXIT_NOSHELL}"
    fi

    # Using `.` instead of `source` as it might not be available for non-bash shells in OSes like Ubuntu,
    # there is no `source` available in `/bin/sh` which is in fact, dash
    docker run --rm --entrypoint /bin/sh "${IMAGE}" -c '. /etc/os-release; echo "${NAME} ${VERSION_ID}"'
    return $?
}
